__version__="4.11.0"
__copyright__="Copyright (c) 2001 - 2026 The SCons Foundation"
__developer__="bdbaddog"
__date__="Mon, 10 Aug 2026 21:23:44 -0700"
__buildsys__="M1Dog2021"
__revision__="190853a12d8c2128f29abe849a6a162f0ed9f049"
__build__="190853a12d8c2128f29abe849a6a162f0ed9f049"
# make sure compatibility is always in place
import SCons.compat  # noqa